"""Contain the Keypair class and related methods"""

from hashlib import sha256

from kin_base.keypair import Keypair as BaseKeypair
from kin_base.exceptions import StellarSecretInvalidError
from kin_base.stellarxdr.StellarXDR_type import DecoratedSignature

from .utils import is_valid_secret_key


class Keypair:
    """A simpler version of kin_base.Keypair that holds the public address and secret seed."""

    def __init__(self, seed=None):
        """
        # Create an instance of Keypair.
        :param seed: (Optional) The secret seed of an account
        """
        self.secret_seed = seed or self.generate_seed()
        if not is_valid_secret_key(self.secret_seed):
            raise StellarSecretInvalidError('invalid seed {}'.format(self.secret_seed))
        base_keypair = BaseKeypair.from_seed(self.secret_seed)
        self.public_address = base_keypair.address().decode()

        # Hint and signing key are needed to sign the tx
        self._hint = base_keypair.signature_hint()
        self._signing_key = base_keypair.signing_key

    def sign(self, data):
        """
        Sign any data using the keypair private key
        :param bytes data: any data to sign
        :return: a decorated signature
        :rtype :class: DecoratedSignature
        """
        signature = self._signing_key.sign(data)
        return DecoratedSignature(self._hint, signature)

    @staticmethod
    def address_from_seed(seed):
        """
        Get a public address from a secret seed.
        :param seed: The secret seed of an account.
        :return: A public address.
        """
        return BaseKeypair.from_seed(seed).address().decode()

    @staticmethod
    def generate_seed():
        """
        Generate a random secret seed.
        :return: A secret seed.
        """
        return BaseKeypair.random().seed().decode()

    @staticmethod
    def generate_hd_seed(base_seed, salt):
        """
        Generate a highly deterministic seed from a base seed + salt
        :param base_seed: The base seed to generate a seed from
        :param salt: A unique string that will be used to generate the seed
        :return: a new seed.
        """
        # Create a new raw seed from the first 32 bytes of this hash
        raw_seed = sha256((base_seed + salt).encode()).digest()[:32]
        return BaseKeypair.from_raw_seed(raw_seed).seed().decode()



"""Contains the KinClient class to interact with the blockchain"""

import requests

from .config import SDK_USER_AGENT, ANON_APP_ID, MAX_RECORDS_PER_REQUEST
from . import errors as KinErrors
from .blockchain.horizon import Horizon
from .monitors import SingleMonitor, MultiMonitor
from .transactions import OperationTypes, SimplifiedTransaction, RawTransaction, build_memo
from .account import KinAccount
from .blockchain.horizon_models import AccountData
from .blockchain.utils import is_valid_address, is_valid_transaction_hash
from .version import __version__

import logging

logger = logging.getLogger(__name__)


class KinClient(object):
    """
    The :class:`kin.KinClient` class is the primary interface to the KIN Python SDK based on Kin Blockchain.
    It maintains a connection context with a Horizon node and hides all the specifics of dealing with Kin REST API.
    """

    def __init__(self, environment):
        """Create a new instance of the KinClient to query the Kin blockchain.
        :param `kin.Environment` environment: an environment for the client to point to.

        :return: An instance of the KinClient.
        :rtype: :class:`KinErrors.KinClient`
        """

        self.environment = environment
        self.network = environment.name

        self.horizon = Horizon(horizon_uri=environment.horizon_uri, user_agent=SDK_USER_AGENT)
        logger.info('Kin SDK inited on network {}, horizon endpoint {}'.format(self.network, self.horizon.horizon_uri))

    def kin_account(self, seed, channel_secret_keys=None, app_id=ANON_APP_ID):
        """
        Create a new instance of a KinAccount to perform authenticated operations on the blockchain.
        :param str seed: The secret seed of the account that will be used
        :param list of str channel_secret_keys: A list of seeds to be used as channels
        :param str app_id: the unique id of your app
        :return: An instance of KinAccount
        :rtype: kin.KinAccount

        :raises: :class:`KinErrors.AccountNotFoundError`: if SDK wallet or channel account is not yet created.
        """

        # Create a new kin account, using self as the KinClient to be used
        return KinAccount(seed, self,channel_secret_keys, app_id)

    def get_config(self):
        """Get system configuration data and online status.
        :return: a dictionary containing the data
        :rtype: dict
        """
        status = {
            'sdk_version': __version__,
            'environment': self.environment.name,
            'horizon': {
                'uri': self.horizon.horizon_uri,
                'online': False,
                'error': None,
            },
            'transport': {
                'pool_size': self.horizon.pool_size,
                'num_retries': self.horizon.num_retries,
                'request_timeout': self.horizon.request_timeout,
                'retry_statuses': self.horizon.status_forcelist,
                'backoff_factor': self.horizon.backoff_factor,
            }
        }

        # now check Horizon connection
        try:
            self.horizon.query('')
            status['horizon']['online'] = True
        except Exception as e:
            status['horizon']['error'] = str(e)

        return status

    def get_minimum_fee(self):
        """
        Get the current minimum fee acceptable for a tx
        :return: The minimum fee
        :type: int
        """
        params = {'order': 'desc',
                  'limit': 1}
        return self.horizon.ledgers(params=params)['_embedded']['records'][0]['base_fee_in_stroops']

    def get_account_balance(self, address):
        """
        Get the KIN balance of a given account
        :param str address: the public address of the account to query
        :return: the balance of the account
        :rtype: float

        :raises: StellarAddressInvalidError: if the provided address has the wrong format.
        :raises: :class:`KinErrors.AccountNotFoundError`: if the account does not exist.
        """

        if not is_valid_address(address):
            raise KinErrors.StellarAddressInvalidError('invalid address: {}'.format(address))
        account_data = self.get_account_data(address)
        for balance in account_data.balances:
            # accounts will always have native asset
            if balance.asset_type == 'native':
                return balance.balance

    def does_account_exists(self, address):
        """
        Find out if a given account exists on the blockchain
        :param str address: The kin account to query about
        :return: does the account exists on the blockchain
        :rtype boolean

        :raises: :class:`KinErrors.StellarAddressInvalidError`: if the address is not valid.
        """

        if not is_valid_address(address):
            raise KinErrors.StellarAddressInvalidError('invalid address: {}'.format(address))

        try:
            self.get_account_balance(address)
            return True
        except KinErrors.AccountNotFoundError:
            return False

    def get_account_data(self, address):
        """Get account data.

        :param str address: the public address of the account to query.

        :return: account data
        :rtype: :class:`kin.AccountData`

        :raises: StellarAddressInvalidError: if the provided address has a wrong format.
        :raises: :class:`KinErrors.AccountNotFoundError`: if the account does not exist.
        """
        # TODO: might want to simplify the returning data
        if not is_valid_address(address):
            raise KinErrors.StellarAddressInvalidError('invalid address: {}'.format(address))

        try:
            acc = self.horizon.account(address)
            return AccountData(acc, strict=False)
        except Exception as e:
            err = KinErrors.translate_error(e)
            raise KinErrors.AccountNotFoundError(address) if \
                isinstance(err, KinErrors.ResourceNotFoundError) else err

    def get_transaction_data(self, tx_hash, simple=True):
        """Gets transaction data.

        :param str tx_hash: transaction hash.
        :param boolean simple: (optional) returns a simplified transaction object

        :return: transaction data
        :rtype: :class:`kin.TransactionData` or `kin.SimplifiedTransaction`

        :raises: ValueError: if the provided hash is invalid.
        :raises: :class:`KinErrors.ResourceNotFoundError`: if the transaction does not exist.
        :raises: :class:`KinErrors.CantSimplifyError`: if the tx is too complex to simplify
        """
        if not is_valid_transaction_hash(tx_hash):
            raise ValueError('invalid transaction hash: {}'.format(tx_hash))

        try:
            raw_tx = RawTransaction(self.horizon.transaction(tx_hash))
        except Exception as e:
            raise KinErrors.translate_error(e)

        if simple:
            return SimplifiedTransaction(raw_tx)
        return raw_tx

    def get_account_tx_history(self, address, amount=10, descending=True, cursor=None, simple=True):
        """
        Get the transaction history for a given account.
        :param str address: The public address of the account to query
        :param int amount: The maximum number of transactions to get
        :param bool descending: The order of the transactions, True will start from the latest one
        :param int cursor: The horizon paging token
        :param bool simple: Should the returned txs be simplified, if True, complicated txs will be ignored
        :return: A list of transactions
        :rtype: list
        """

        if not is_valid_address(address):
            raise KinErrors.StellarAddressInvalidError('invalid address: {}'.format(address))

        if amount <= 0:
            raise ValueError('Limit must be bigger than 0')

        tx_list = []

        requested_amount = amount if amount < MAX_RECORDS_PER_REQUEST else MAX_RECORDS_PER_REQUEST
        params = {
            'limit': requested_amount,
            'order': 'desc' if descending else 'asc'
        }

        # cursor is optional
        if cursor is not None:
            params['cursor'] = cursor

        horizon_response = self.horizon.account_transactions(address, params)

        for transaction in horizon_response['_embedded']['records']:
            raw_tx = RawTransaction(transaction)
            if simple:
                try:
                    simple_tx = SimplifiedTransaction(raw_tx)
                    tx_list.append(simple_tx)
                except KinErrors.CantSimplifyError:
                    pass
            else:
                tx_list.append(raw_tx)
            last_cursor = transaction['paging_token']

        remaining_txs = amount - len(tx_list)
        # if we got all the txs that we wanted, or there are no more txs
        if remaining_txs <= 0 or len(horizon_response['_embedded']['records']) < amount:
            return tx_list
        # If there are anymore transactions, recursively get the next transaction page
        return tx_list.extend(self.get_account_tx_history(address, remaining_txs, descending, last_cursor, simple))

    def verify_kin_payment(self, tx_hash, source, destination, amount, memo=None, check_memo=False, app_id=ANON_APP_ID):
        """
        Verify that a give tx matches the desired parameters
        :param str tx_hash: The hash of the transaction to query
        :param str source: The expected source account
        :param str destination: The expected destination account
        :param float amount: The expected amount
        :param str memo: (optional) The expected memo
        :param boolean check_memo: (optional) Should the memo match
        :param str app_id: the id of the app that sent the tx
        :return: True/False
        :rtype: boolean
        """

        try:
            tx = self.get_transaction_data(tx_hash)
            operation = tx.operation
            if operation.type != OperationTypes.PAYMENT:
                return False
            if source != tx.source or destination != operation.destination or amount != operation.amount:
                return False
            if check_memo and build_memo(app_id, memo) != tx.memo:
                return False

            return True

        except KinErrors.CantSimplifyError:
            return False

    def friendbot(self, address):
        """
        Use the friendbot service to create and fund an account
        :param str address: The address to create and fund
        :return: the hash of the friendobt transaction
        :rtype str

        :raises ValueError: if no friendbot service was provided
        :raises ValueError: if the address is invalid
        :raises :class: `KinErrors.AccountExistsError`: if the account already exists
        :raises :class: `KinErrors.FriendbotError`: If the friendbot request failed
        """

        if self.environment.friendbot_url is None:
            raise ValueError("No friendbot service was configured for this client's environments")

        if not is_valid_address(address):
            raise KinErrors.StellarAddressInvalidError('invalid address: {}'.format(address))
        if self.does_account_exists(address):
            raise KinErrors.AccountExistsError(address)

        response = requests.get(self.environment.friendbot_url, params={'addr': address})
        if response.ok:
            return response.json()['hash']
        else:
            raise KinErrors.FriendbotError(response.status_code, response.text)

    def monitor_account_payments(self, address, callback_fn):
        """Monitor KIN payment transactions related to the account identified by provided address.
        NOTE: the function starts a background thread.

        :param str address: the address of the account to query.

        :param callback_fn: the function to call on each received payment as `callback_fn(address, tx_data, monitor)`.
        :type: callable[str,:class:`kin.TransactionData`,:class:`kin.SingleMonitor`]

        :return: a monitor instance
        :rtype: :class:`kin.SingleMonitor`

        :raises: ValueError: when no address is given.
        :raises: ValueError: if the address is in the wrong format
        :raises: :class:`KinErrors.AccountNotActivatedError`: if the account given is not activated
        """

        return SingleMonitor(self, address, callback_fn)

    def monitor_accounts_payments(self, addresses, callback_fn):
        """Monitor KIN payment transactions related to multiple accounts
        NOTE: the function starts a background thread.

        :param str addresses: the addresses of the accounts to query.

        :param callback_fn: the function to call on each received payment as `callback_fn(address, tx_data, monitor)`.
        :type: callable[str,:class:`kin.TransactionData`,:class:`kin.MultiMonitor`]

        :return: a monitor instance
        :rtype: :class:`kin.MultiMonitor`

        :raises: ValueError: when no address is given.
        :raises: ValueError: if the addresses are in the wrong format
        :raises: :class:`KinErrors.AccountNotActivatedError`: if the accounts given are not activated
        """

        return MultiMonitor(self, addresses, callback_fn)



"""Contains the config for the Kin SDK"""

from .blockchain.environment import Environment

from .version import __version__

# Create default environments.
HORIZON_URI_PROD = 'https://horizon-ecosystem.kininfrastructure.com'
HORIZON_PASSPHRASE_PROD = 'Public Global Kin Ecosystem Network ; June 2018'
PROD_ENVIRONMENT = Environment('ECOSYSTEM', HORIZON_URI_PROD, HORIZON_PASSPHRASE_PROD)

HORIZON_URI_TEST = 'https://horizon-playground.kininfrastructure.com'
HORIZON_PASSPHRASE_TEST = 'Kin Playground Network ; June 2018'
FRIENDBOT_URL_TEST = "http://friendbot-playground.kininfrastructure.com"
TEST_ENVIRONMENT = Environment('PLAYGROUND', HORIZON_URI_TEST, HORIZON_PASSPHRASE_TEST, FRIENDBOT_URL_TEST)

MEMO_CAP = 28
# Template version - app_id - free text
MEMO_TEMPLATE = '1-{}-'
# hardcoded id for anonymous users
ANON_APP_ID = 'anon'
# 4 letters/numbers
APP_ID_REGEX = '^[a-zA-Z0-9]{4}$'

MAX_RECORDS_PER_REQUEST = 200

SDK_USER_AGENT = 'kin-core-python/{}'.format(__version__)



"""Contains the channel manager class to take care of channels"""

import sys

from kin_base.keypair import Keypair

from .builder import Builder
from .errors import ChannelsBusyError

import logging

logger = logging.getLogger(__name__)

if sys.version[0] == '2':
    import Queue as queue
else:
    # noinspection PyUnresolvedReferences
    import queue as queue

CHANNEL_QUEUE_TIMEOUT = 11  # how much time to wait until a channel is available, in seconds
CHANNEL_PUT_TIMEOUT = 0.5  # how much time to wait for a channel to return to the queue


class ChannelManager(object):
    """ The class :class:`kin.ChannelManager` wraps channel-related specifics of transaction sending."""

    def __init__(self, secret_key, channel_keys, network, horizon):
        self.base_key = secret_key
        self.base_address = Keypair.from_seed(secret_key).address().decode()
        self.num_channels = len(channel_keys)
        self.channel_builders = queue.Queue(len(channel_keys))
        self.horizon = horizon
        self.low_balance_builders = []
        for channel_key in channel_keys:
            # create a channel transaction builder.
            # fee gets updated once a transaction is being built
            builder = Builder(secret=channel_key, network=network, horizon=horizon, fee=0.01)
            self.channel_builders.put(builder)

    def build_transaction(self, add_ops_fn, fee, memo_text=None):
        """Send a transaction using an available channel account.

        :param add_ops_fn: a function to call, that will add operations to the transaction. The function should be
            `partial`, because a `source` parameter will be added.
        :type add_ops_fn: callable[builder]

        :param str memo_text: (optional) a text to add as transaction memo.

        :return: transaction object
        :rtype: dict
        """
        # get an available channel builder first (blocking with timeout)
        try:
            builder = self.channel_builders.get(True, CHANNEL_QUEUE_TIMEOUT)
        except queue.Empty:
            raise ChannelsBusyError

        # operation source is always the base account
        source = self.base_address if builder.address != self.base_address else None

        # add operation (using external partial) and sign
        try:
            add_ops_fn(builder)(source=source)
            # update the previous fee that was only used for initialization
            builder.fee = fee
            if memo_text:
                builder.add_text_memo(memo_text)  # max memo length is 28

            builder.sign()  # always sign with a channel key
            if source:
                builder.sign(secret=self.base_key)  # sign with the base key if needed
        except:
            # If something fails when building the tx, clear and return the builder to queue.
            builder.clear()
            self.channel_builders.put(builder,timeout=CHANNEL_PUT_TIMEOUT)
            raise
        return builder



"""Contains data models related to horizon"""

from schematics.models import Model
from schematics.types import IntType, BooleanType, StringType, UTCDateTimeType, FloatType
from schematics.types.compound import ModelType, ListType, DictType


class PModel(Model):
    """Base class for our models that provides printout capabilities"""

    def __str__(self):
        sb = []
        for key in self.__dict__:
            if not key.startswith('__'):
                sb.append("\t{}='{}'".format(key, self.__dict__[key]))
        return '\n'.join(sb)

    def __repr__(self):
        return self.__str__()

    def __hash__(self):
        return hash(self.__str__())


class AccountData(PModel):
    class Thresholds(PModel):
        low_threshold = IntType(default=0)
        medium_threshold = IntType(default=0)
        high_threshold = IntType(default=0)

    class Flags(PModel):
        """Flags set on issuer accounts.
           TrustLines are created with authorized set to "false" requiring
           the issuer to set it for each TrustLine
        """
        auth_required = BooleanType(default=False)  # If set, the authorized flag in TrustLines can be cleared.
        # Otherwise, authorization cannot be revoked
        auth_revocable = BooleanType(default=False)  # Once set, causes all AUTH_* flags to be read-only

    class Balance(PModel):
        asset_type = StringType()
        asset_code = StringType()
        asset_issuer = StringType()
        balance = FloatType(default=0)
        limit = FloatType()

    class Signer(PModel):
        public_key = StringType()
        key = StringType()
        weight = IntType()
        signature_type = StringType(serialized_name='type')

    id = StringType()
    account_id = StringType()
    sequence = StringType()
    data = DictType(StringType, default={})
    thresholds = ModelType(Thresholds)
    balances = ListType(ModelType(Balance), default=[])
    flags = ModelType(Flags)
    paging_token = StringType()
    subentry_count = IntType()
    signers = ListType(ModelType(Signer), default=[])


class TransactionResultCodes(PModel):
    transaction = StringType()
    operations = ListType(StringType, default=[])


class HTTPProblemDetails(PModel):
    """HTTP Problem Details object.
    See https://tools.ietf.org/html/rfc7807
    """

    class Extras(PModel):
        invalid_field = StringType()
        envelope_xdr = StringType()
        result_xdr = StringType()
        result_codes = ModelType(TransactionResultCodes)

    type = StringType()
    title = StringType()
    status = IntType()
    detail = StringType()
    instance = StringType()
    extras = ModelType(Extras)



"""Contains the builder class to build transactions"""

from kin_base.builder import Builder as BaseBuilder
from kin_base.keypair import Keypair
from kin_base.memo import NoneMemo
from kin_base.exceptions import StellarAddressInvalidError

from .utils import is_valid_address, is_valid_secret_key


class Builder(BaseBuilder):
    """
    This class overrides :class:`kin_base.builder` to provide additional functionality.
    """

    def __init__(self, network, horizon, fee, secret=None, address=None):
        if secret:
            if not is_valid_secret_key(secret):
                raise ValueError('invalid secret key')
            address = Keypair.from_seed(secret).address().decode()
        elif address:
            if not is_valid_address(address):
                raise StellarAddressInvalidError('invalid address: {}'.format(address))
        else:
            raise Exception('either secret or address must be provided')

        # call baseclass constructor to init base class variables
        super(Builder, self).__init__(secret=secret, address=address, sequence=1, fee=fee)

        # custom overrides

        self.network = network
        self.horizon = horizon

    def clear(self):
        """"Clears the builder so it can be reused."""
        self.ops = []
        self.time_bounds = None
        self.memo = NoneMemo()
        self.tx = None
        self.te = None

    def get_sequence(self):
        """Alternative implementation to expose exceptions"""
        return self.horizon.account(self.address).get('sequence')

    def next(self):
        """
        Alternative implementation that does not create a new builder but clears the current one and increments
        the account sequence number.
        """
        self.clear()
        self.sequence = str(int(self.sequence) + 1)

    def sign(self, secret=None):
        """
        Alternative implementation that does not use the self-managed sequence, but always fetches it from Horizon.
        """
        if not secret:  # only get the new sequence for my own account
            self.sequence = self.get_sequence()
        super(Builder, self).sign(secret)



"""Contains the KinAccount and AccountStatus classes."""

import re
from functools import partial
import json

from kin_base.transaction_envelope import TransactionEnvelope
from kin_base.network import NETWORKS

from .blockchain.keypair import Keypair
from .blockchain.horizon import Horizon
from .blockchain.builder import Builder
from .blockchain.channel_manager import ChannelManager
from . import errors as KinErrors
from .transactions import Transaction, build_memo
from .blockchain.errors import TransactionResultCode, HorizonErrorType, HorizonError
from .config import SDK_USER_AGENT, APP_ID_REGEX
from .blockchain.utils import is_valid_secret_key, is_valid_address

import logging

logger = logging.getLogger(__name__)


class KinAccount:
    """Account class to perform authenticated actions on the blockchain"""

    def __init__(self, seed, client, channel_secret_keys, app_id):
        # Set the internal sdk
        self._client = client

        # Set the app_id
        self.app_id = app_id

        # Verify the app_id is ok
        if re.match(APP_ID_REGEX, app_id) is None:
            raise ValueError('invalid app id: {}'.format(app_id))

        # Verify seed
        if not is_valid_secret_key(seed):
            raise ValueError('invalid secret key: {}'.format(seed))

        # Set keypair
        self.keypair = Keypair(seed)
        # check that sdk wallet account exists
        if not self._client.does_account_exists(self.keypair.public_address):
            raise KinErrors.AccountNotFoundError(self.keypair.public_address)

        if channel_secret_keys is not None:
            # Use given channels
            self.channel_secret_keys = channel_secret_keys
        else:
            # Use the base account as the only channel
            self.channel_secret_keys = [seed]

        for channel_key in self.channel_secret_keys:
            # Verify channel seed
            if not is_valid_secret_key(channel_key):
                raise KinErrors.StellarSecretInvalidError('invalid channel key: {}'.format(channel_key))
            # Check that channel accounts exists.
            channel_address = Keypair.address_from_seed(channel_key)
            if not self._client.does_account_exists(channel_address):
                raise KinErrors.AccountNotFoundError(channel_address)

        # set connection pool size for channels + monitoring connection + extra
        pool_size = max(1, len(self.channel_secret_keys)) + 2

        # Set an horizon instance with the new pool_size
        self.horizon = Horizon(self._client.environment.horizon_uri,
                               pool_size=pool_size, user_agent=SDK_USER_AGENT)
        self.channel_manager = ChannelManager(seed, self.channel_secret_keys,
                                              self._client.environment.name, self.horizon)

    def get_public_address(self):
        """Return this KinAccount's public address"""
        return self.keypair.public_address

    def get_balance(self):
        """
        Get the KIN balance of this KinAccount
        :return: the kin balance
        :rtype: float

        :raises: :class:`KinErrors.AccountNotFoundError`: if the account does not exist.
        """
        return self._client.get_account_balance(self.keypair.public_address)

    def get_data(self):
        """
        Gets this KinAccount's data

        :return: account data
        :rtype: :class:`kin.AccountData`

        :raises: :class:`KinErrors.AccountNotFoundError`: if the account does not exist.
        """
        return self._client.get_account_data(self.keypair.public_address)

    def get_transaction_history(self, amount=10, descending=True, cursor=None, simple=True):
        """
        Get the transaction history for this kin account
        :param int amount: The maximum number of transactions to get
        :param bool descending: The order of the transactions, True will start from the latest one
        :param int cursor: The horizon paging token
        :param bool simple: Should the returned txs be simplified, if True, complicated txs will be ignored
        :return: A list of transactions
        :rtype: list
        """

        return self._client.get_account_tx_history(self.get_public_address(),
                                                   amount=amount,
                                                   descending=descending,
                                                   cursor=cursor,
                                                   simple=simple)

    def create_account(self, address, starting_balance, fee, memo_text=None):
        """Create an account identified by the provided address.

        :param str address: the address of the account to create.

        :param number starting_balance: the starting KIN balance of the account.

        :param str memo_text: (optional) a text to put into transaction memo, up to MEMO_CAP chars.

        :param float fee: fee to be deducted for the tx

        :return: the hash of the transaction
        :rtype: str

        :raises: ValueError: if the supplied address has a wrong format.
        :raises: :class:`KinErrors.AccountExistsError`: if the account already exists.
        :raises: :class:`KinErrors.NotValidParamError`: if the memo is longer than MEMO_CAP characters
        :raises: KinErrors.NotValidParamError: if the amount is too precise
        :raises: KinErrors.NotValidParamError: if the fee is not an too precise
        """
        tx = self.build_create_account(address,
                                       starting_balance=starting_balance,
                                       fee=fee,
                                       memo_text=memo_text)
        return self.submit_transaction(tx)

    def send_kin(self, address, amount, fee, memo_text=None):
        """Send KIN to the account identified by the provided address.

        :param str address: the account to send KIN to.

        :param number amount: the amount of KIN to send.

        :param str memo_text: (optional) a text to put into transaction memo.

        :param float fee: fee to be deducted

        :return: the hash of the transaction
        :rtype: str

        :raises: ValueError: if the provided address has a wrong format.
        :raises: ValueError: if the amount is not positive.
        :raises: KinErrors.NotValidParamError: if the amount is too precise
        :raises: :class:`KinErrors.AccountNotFoundError`: if the account does not exist.
        :raises: :class:`KinErrors.AccountNotActivatedError`: if the account is not activated.
        :raises: :class:`KinErrors.LowBalanceError`: if there is not enough KIN and XLM to send and pay transaction fee.
        :raises: :class:`KinErrors.NotValidParamError`: if the memo is longer than MEMO_CAP characters
        :raises: KinErrors.NotValidParamError: if the fee is not an too precise
        """
        tx = self.build_send_kin(address, amount, fee, memo_text)
        return self.submit_transaction(tx)

    def build_create_account(self, address, starting_balance, fee, memo_text=None):
        """Build a tx that will create an account identified by the provided address.

        :param str address: the address of the account to create.

        :param number starting_balance: the starting XLM balance of the account.

        :param str memo_text: (optional) a text to put into transaction memo, up to MEMO_CAP chars.

        :param float fee: fee to be deducted for the tx

        :return: a transaction object
        :rtype: :class: `Kin.Transaction`

        :raises: ValueError: if the supplied address has a wrong format.
        :raises: :class:`KinErrors.NotValidParamError`: if the memo is longer than MEMO_CAP characters
        :raises: KinErrors.NotValidParamError: if the amount is too precise
        :raises: KinErrors.NotValidParamError: if the fee is not an too precise
        """
        if not is_valid_address(address):
            raise KinErrors.StellarAddressInvalidError('invalid address: {}'.format(address))

        if float(starting_balance) < 0:
            raise ValueError('Starting balance : {} cant be negative'.format(starting_balance))

        # Build the transaction and send it.

        builder = self.channel_manager.build_transaction(lambda builder:
                                                         partial(builder.append_create_account_op, address,
                                                                 str(starting_balance)),
                                                         fee,
                                                         memo_text=build_memo(self.app_id, memo_text))
        return Transaction(builder, self.channel_manager)

    def build_send_kin(self, address, amount, fee, memo_text=None):
        """Build a tx to send KIN to the account identified by the provided address.

        :param str address: the account to send asset to.

        :param number amount: the asset amount to send.

        :param str memo_text: (optional) a text to put into transaction memo.

        :param float fee: fee to be deducted for the tx

        :return: a transaction object
        :rtype: :class: `Kin.Transaction`

        :raises: ValueError: if the provided address has a wrong format.
        :raises: ValueError: if the amount is not positive.
        :raises: KinErrors.NotValidParamError: if the amount is too precise
        :raises: KinErrors.NotValidParamError: if the fee is not an too precise
        """

        if not is_valid_address(address):
            raise KinErrors.StellarAddressInvalidError('invalid address: {}'.format(address))

        if float(amount) <= 0:
            raise ValueError('Amount : {} must be positive'.format(amount))

        builder = self.channel_manager.build_transaction(lambda builder:
                                                         partial(builder.append_payment_op, address, str(amount)),
                                                         fee,
                                                         memo_text=build_memo(self.app_id, memo_text))
        return Transaction(builder, self.channel_manager)

    def submit_transaction(self, tx, is_re_submitting=False):
        """
        Submit a transaction to the blockchain.
        :param :class: `kin.Transaction` tx: The transaction object to send
        :param boolean is_re_submitting: is this a re-submission
        :return: The hash of the transaction.
        :rtype: str
        """
        try:
            return tx.builder.submit()['hash']
        # If the channel is out of KIN, top it up and try again
        except HorizonError as e:
            logging.warning('send transaction error with channel {}: {}'.format(tx.builder.address, str(e)))
            if e.type == HorizonErrorType.TRANSACTION_FAILED \
                    and e.extras.result_codes.transaction == TransactionResultCode.INSUFFICIENT_BALANCE:
                tx.channel_manager.low_balance_builders.append(tx.builder)
                self._top_up(tx.builder.address)
                tx.channel_manager.low_balance_builders.remove(tx.builder)

                # Insufficient balance is a "fast-fail", the sequence number doesn't increment
                # so there is no need to build the transaction again
                self.submit_transaction(tx, is_re_submitting=True)
            else:
                raise KinErrors.translate_error(e)
        finally:
            if not is_re_submitting:
                tx.release()

    def monitor_payments(self, callback_fn):
        """Monitor KIN payment transactions related to this account
        NOTE: the function starts a background thread.

        :param callback_fn: the function to call on each received payment as `callback_fn(address, tx_data, monitor)`.
        :type: callable[str,:class:`kin.TransactionData`,:class:`kin.SingleMonitor`]

        :return: a monitor instance
        :rtype: :class:`kin.SingleMonitor`
        """
        return self._client.monitor_account_payments(self.keypair.public_address, callback_fn)

    def whitelist_transaction(self, payload):
        """
        Sign on a transaction to whitelist it
        :param str payload: the json received from the client
        :return: a signed transaction encoded as base64
        :rtype str
        """

        # load the object from the json
        if not isinstance(payload, dict):
            payload = json.loads(payload)

        # If the network the client is using is different from the one we are using
        if NETWORKS[self._client.environment.name] != payload['network_id']:
            raise KinErrors.WrongNetworkError()

        # The android stellar sdk spells 'tx-envelope' as 'envelop'
        payload_envelope = payload.get('envelop',payload.get('envelope'))
        # Decode the transaction, from_xdr actually takes a base64 encoded xdr
        envelope = TransactionEnvelope.from_xdr(payload_envelope)

        # Add the network_id hash to the envelop
        envelope.network_id = self._client.environment.passphrase_hash

        # Get the transaction hash (to sign it)
        tx_hash = envelope.hash_meta()

        # Sign using the hash
        signature = self.keypair.sign(tx_hash)

        # Add the signature to the envelop
        envelope.signatures.append(signature)

        # Pack the signed envelop to xdr the return it encoded as base64
        # xdr() returns a bytestring that needs to be decoded
        return envelope.xdr().decode()


    # Internal methods

    def _top_up(self, address):
        """
        Top up a channel with the base account.
        :param str address: The address to top up
        """
        # In theory, if the sdk runs in threads, and 2 or more channels
        # are out of funds and needed to be topped up at the exact same time
        # there is a chance for a bad_sequence error,
        # however it is virtually impossible that this situation will occur.

        # TODO: let user config the amount of kin to top up
        builder = Builder(self._client.environment.name, self._client.horizon,
                          self._client.get_minimum_fee(), self.keypair.secret_seed)
        builder.append_payment_op(address, str(self._client.get_minimum_fee() * 1000))
        builder.sign()
        builder.submit()



"""Contains the Environment class to configure horizon"""

from hashlib import sha256

from kin_base.network import NETWORKS
from kin_base.asset import Asset
from kin_base.exceptions import StellarAddressInvalidError

from .utils import is_valid_address


class Environment:
    """Environments holds the parameters that will be used to connect to horizon"""
    def __init__(self, name, horizon_endpoint_uri, network_passphrase, friendbot_url=None):
        """

        :param str name: Name of the environment.
        :param str horizon_uri: a Horizon endpoint.
        :param str network_passphrase: The passphrase/network_id of the environment.
        :param str friendbot_url: a url to a friendbot service
        :return: An instance of the Environment class.
        :rtype: :class: `kin.Environment`

        :raises: ValueError: is the kin_issuer is invalid
        """
        # Add the network to the kin_base network list.
        NETWORKS[name.upper()] = network_passphrase
        self.name = name.upper()
        self.horizon_uri = horizon_endpoint_uri
        self.friendbot_url = friendbot_url

        # Calculate the hash of the passphrase, can be used to calculate tx hash.
        self.passphrase_hash = sha256(network_passphrase.encode()).digest()



"""Contains helper methods for the kin sdk"""

from hashlib import sha256

from .client import KinClient
from .blockchain.builder import Builder
from .blockchain.keypair import Keypair
from .errors import AccountNotFoundError


def create_channels(master_seed, environment, amount, starting_balance, salt):
    """
    Create HD seeds based on a master seed and salt
    :param str master_seed: The master seed that creates the seeds
    :param Kin.Environment environment: The blockchain environment to create the seeds on
    :param int amount: Number of seeds to create (Up to 100)
    :param float starting_balance: Starting balance to create channels with
    :param str salt: A string to be used to create the HD seeds
    :return: The list of seeds generated
    :rtype list[str]
    """

    client = KinClient(environment)
    base_key = Keypair(master_seed)
    if not client.does_account_exists(base_key.public_address):
        raise AccountNotFoundError(base_key.public_address)

    fee = client.get_minimum_fee()

    channels = get_hd_channels(master_seed, salt, amount)

    # Create a builder for the transaction
    builder = Builder(environment.name, client.horizon, fee, master_seed)

    # Find out if this salt+seed combination was ever used to create channels.
    # If so, the user might only be interested in adding channels,
    # so we need to find what seed to start from

    for index, seed in enumerate(channels):
        if client.does_account_exists(Keypair.address_from_seed(seed)):
            continue

        # Start creating from the current seed forward
        for channel_seed in channels[index:]:
            builder.append_create_account_op(Keypair.address_from_seed(channel_seed), str(starting_balance))

        builder.sign()
        builder.submit()
        break

    return channels


def get_hd_channels(master_seed, salt, amount):
    """
    Get a list of channels generated based on a seed and salt
    :param str master_seed: the base seed that created the channels
    :param str salt: A string to be used to generate the seeds
    :param int amount: Number of seeds to generate (Up to 100)
    :return: The list of seeds generated
    :rtype list[str]
    """

    if amount > 100:
        raise ValueError('Only 100 channels can be created with a specific seed + salt combination')
    hashed_salt = sha256(salt.encode()).hexdigest()

    channels = []
    for index in range(amount):
        # The salt used to generate the
        channel = Keypair.generate_hd_seed(master_seed, hashed_salt + str(index))
        channels.append(channel)

    return channels







"""Contains the Horizon class to interact with horizon"""

import requests
from requests.adapters import HTTPAdapter, DEFAULT_POOLSIZE
from requests.exceptions import RequestException
import sys
from time import sleep
from urllib3.util import Retry

from kin_base.horizon import HORIZON_LIVE, HORIZON_TEST

from .errors import HorizonError

import logging

logger = logging.getLogger(__name__)

try:
    from sseclient import SSEClient
except ImportError:
    SSEClient = None

if sys.version[0] == '2':
    # noinspection PyUnresolvedReferences
    from urllib import urlencode
else:
    # noinspection PyUnresolvedReferences
    from urllib.parse import urlencode

DEFAULT_REQUEST_TIMEOUT = 11  # two ledgers + 1 sec, let's retry faster and not wait 60 secs.
DEFAULT_NUM_RETRIES = 5
DEFAULT_BACKOFF_FACTOR = 0.5
USER_AGENT = 'py-stellar-base'


class Horizon(object):
    """
    This class redefines :class:`kin_base.horizon.Horizon` to provide additional functionality:
        - persistent connection to Horizon and connection pool
        - configurable request retry functionality
        - Horizon error checking and deserialization
    """

    def __init__(self, horizon_uri=None, pool_size=DEFAULT_POOLSIZE, num_retries=DEFAULT_NUM_RETRIES,
                 request_timeout=DEFAULT_REQUEST_TIMEOUT, backoff_factor=DEFAULT_BACKOFF_FACTOR, user_agent=USER_AGENT):
        if horizon_uri is None:
            self.horizon_uri = HORIZON_TEST
        else:
            self.horizon_uri = horizon_uri

        self.pool_size = pool_size
        self.num_retries = num_retries
        self.request_timeout = request_timeout
        self.backoff_factor = backoff_factor

        # adding 504 to the list of statuses to retry
        self.status_forcelist = list(Retry.RETRY_AFTER_STATUS_CODES)
        self.status_forcelist.append(504)

        # configure standard session

        # configure retry handler
        retry = Retry(total=self.num_retries, backoff_factor=self.backoff_factor, redirect=0,
                      status_forcelist=self.status_forcelist)
        # init transport adapter
        adapter = HTTPAdapter(pool_connections=self.pool_size, pool_maxsize=self.pool_size, max_retries=retry)

        # init session
        session = requests.Session()

        # set default headers
        session.headers.update({'User-Agent': user_agent})

        session.mount('http://', adapter)
        session.mount('https://', adapter)
        self._session = session

        # configure SSE session (differs from our standard session)

        sse_retry = Retry(total=1000000, redirect=0, status_forcelist=self.status_forcelist)
        sse_adapter = HTTPAdapter(pool_connections=self.pool_size, pool_maxsize=self.pool_size, max_retries=sse_retry)
        sse_session = requests.Session()
        sse_session.headers.update({'User-Agent': user_agent})
        sse_session.mount('http://', sse_adapter)
        sse_session.mount('https://', sse_adapter)
        self._sse_session = sse_session

    def submit(self, te):
        """Submit the transaction using a pooled connection, and retry on failure."""
        params = {'tx': te}
        url = self.horizon_uri + '/transactions/'

        # POST is not included in Retry's method_whitelist for a good reason.
        # our custom retry mechanism follows
        reply = None
        retry_count = self.num_retries
        while True:
            try:
                reply = self._session.post(url, data=params, timeout=self.request_timeout)
                return check_horizon_reply(reply.json())
            except (RequestException, ValueError) as e:
                if reply:
                    msg = 'horizon submit exception: {}, reply: [{}] {}'.format(str(e), reply.status_code, reply.text)
                else:
                    msg = 'horizon submit exception: {}'.format(str(e))
                logging.warning(msg)

                if reply and reply.status_code not in self.status_forcelist:
                    raise Exception('invalid horizon reply: [{}] {}'.format(reply.status_code, reply.text))
                # retry
                if retry_count <= 0:
                    raise
                retry_count -= 1
                logging.warning('submit retry attempt {}'.format(retry_count))
                sleep(self.backoff_factor)

    def query(self, rel_url, params=None, sse=False):
        abs_url = self.horizon_uri + rel_url
        reply = self._query(abs_url, params, sse)
        return check_horizon_reply(reply) if not sse else reply

    def account(self, address):
        url = '/accounts/' + address
        return self.query(url)

    def account_effects(self, address, params=None, sse=False):
        url = '/accounts/' + address + '/effects/'
        return self.query(url, params, sse)

    def account_offers(self, address, params=None):
        url = '/accounts/' + address + '/offers/'
        return self.query(url, params)

    def account_operations(self, address, params=None, sse=False):
        url = '/accounts/' + address + '/operations/'
        return self.query(url, params, sse)

    def account_transactions(self, address, params=None, sse=False):
        url = '/accounts/' + address + '/transactions/'
        return self.query(url, params, sse)

    def account_payments(self, address, params=None, sse=False):
        url = '/accounts/' + address + '/payments/'
        return self.query(url, params, sse)

    def transactions(self, params=None, sse=False):
        url = '/transactions/'
        return self.query(url, params, sse)

    def transaction(self, tx_hash):
        url = '/transactions/' + tx_hash
        return self.query(url)

    def transaction_operations(self, tx_hash, params=None):
        url = '/transactions/' + tx_hash + '/operations/'
        return self.query(url, params)

    def transaction_effects(self, tx_hash, params=None):
        url = '/transactions/' + tx_hash + '/effects/'
        return self.query(url, params)

    def transaction_payments(self, tx_hash, params=None):
        url = '/transactions/' + tx_hash + '/payments/'
        return self.query(url, params)

    def order_book(self, params=None):
        url = '/order_book/'
        return self.query(url, params)

    def trades(self, params=None):
        url = '/trades/'
        return self.query(url, params)

    def ledgers(self, params=None, sse=False):
        url = '/ledgers/'
        return self.query(url, params, sse)

    def ledger(self, ledger_id):
        url = '/ledgers/' + str(ledger_id)
        return self.query(url)

    def ledger_transactions(self, ledger_id, params=None):
        url = '/ledgers/' + str(ledger_id) + '/transactions/'
        return self.query(url, params)

    def ledger_effects(self, ledger_id, params=None):
        url = '/ledgers/' + str(ledger_id) + '/effects/'
        return self.query(url, params)

    def ledger_operations(self, ledger_id, params=None):
        url = '/ledgers/' + str(ledger_id) + '/operations/'
        return self.query(url, params)

    def ledger_payments(self, ledger_id, params=None):
        url = '/ledgers/' + str(ledger_id) + '/payments/'
        return self.query(url, params)

    def effects(self, params=None, sse=False):
        url = '/effects/'
        return self.query(url, params, sse)

    def operations(self, params=None, sse=False):
        url = '/operations/'
        return self.query(url, params, sse)

    def operation(self, op_id, params=None):
        url = '/operations/' + str(op_id)
        return self.query(url, params)

    def operation_effects(self, op_id, params=None):
        url = '/operations/' + str(op_id) + '/effects/'
        return self.query(url, params)

    def payments(self, params=None, sse=False):
        url = '/payments/'
        return self.query(url, params, sse)

    def assets(self, params=None):
        url = '/assets/'
        return self.query(url, params)

    def _query(self, url, params=None, sse=False):
        if not sse:
            reply = self._session.get(url, params=params, timeout=self.request_timeout)
            try:
                return reply.json()
            except ValueError:
                raise Exception('invalid horizon reply: [{}] {}'.format(reply.status_code, reply.text))

        # SSE connection
        if SSEClient is None:
            raise ValueError('SSE not supported, missing sseclient module')

        return SSEClient(url, session=self._sse_session, params=params)

    @staticmethod
    def testnet():
        return Horizon(horizon_uri=HORIZON_TEST)

    @staticmethod
    def livenet():
        return Horizon(horizon_uri=HORIZON_LIVE)


def check_horizon_reply(reply):
    if 'status' not in reply:
        return reply
    raise HorizonError(reply)



#!/usr/bin/env python

from setuptools import setup, find_packages

exec(open("kin/version.py").read())

with open('requirements.txt') as f:
    requires = [line.split(' ')[0] for line in f]
with open('requirements-dev.txt') as f:
    tests_requires = [line.split(' ')[0] for line in f]

setup(
    name='kin-sdk',
    version=__version__,
    description='KIN SDK for Python',
    author='Kin Ecosystem',
    author_email='david.bolshoy@kik.com',
    maintainer='Ron Serruya',
    maintainer_email='ron.serruya@kik.com',
    url='https://github.com/kinecosystem/kin-core-python/tree/v2-master',
    license='MIT',
    packages=find_packages(),
    long_description=open("README.md").read(),
    keywords=["kin", "stellar", "blockchain", "cryptocurrency"],
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
    ],
    install_requires=requires,
    tests_require=tests_requires,
    python_requires='>=2.7',
)



"""Contains classes for monitoring the blockchain"""
from threading import Thread, Event

from .blockchain.utils import is_valid_address
from .transactions import OperationTypes, SimplifiedTransaction ,RawTransaction
from .errors import AccountNotFoundError, CantSimplifyError, StoppedMonitorError, StellarAddressInvalidError

import logging

logger = logging.getLogger(__name__)


class SingleMonitor:
    """Single Monitor to monitor Kin payment on a single account"""

    def __init__(self, kin_client, address, callback_fn):
        """
        Monitors a single account for kin payments
        :param kin_client: a kin client directed to the correct network
        :param address: address to watch
        :param callback_fn: function to callback when the payment is found
        """
        self.kin_client = kin_client
        self.callback_fn = callback_fn

        if not address:
            raise ValueError('no address to monitor')

        if not is_valid_address(address):
            raise StellarAddressInvalidError('invalid address: {}'.format(address))

        if not self.kin_client.does_account_exists(address):
            raise AccountNotFoundError(address)

        self.address = address

        # Currently, due to nonstandard SSE implementation in Horizon,
        # using cursor=now will block until the first tx happens.
        # Instead, we determine the cursor ourselves.
        # Fix will be for horizon to send any message just to start a connection.
        # This will cause a tx
        params = {}
        reply = self.kin_client.horizon.account_transactions(address, params={'order': 'desc', 'limit': 2})
        if len(reply['_embedded']['records']) == 2:
            cursor = reply['_embedded']['records'][1]['paging_token']
            params = {'cursor': cursor}

        # make synchronous SSE request (will raise errors in the current thread)
        self.sse_client = self.kin_client.horizon.account_transactions(address, sse=True, params=params)

        self.stop_event = Event()
        # start monitoring thread
        self.thread = Thread(target=self.event_processor, args=(self.stop_event,))
        self.thread.daemon = True
        self.thread.start()

    def event_processor(self, stop_event):
        """
        Method to filter through SSE events and find kin payments for an account
        :param stop_event: an event that can be used to stop this method
        """
        import json
        try:
            for event in self.sse_client:
                if stop_event.is_set():
                    return
                if event.event != 'message':
                    continue
                try:
                    tx = json.loads(event.data)

                    try:
                        tx_data = SimplifiedTransaction(RawTransaction(tx))
                    except CantSimplifyError:
                        continue

                    if tx_data.operation.type != OperationTypes.PAYMENT:
                        continue

                    self.callback_fn(self.address, tx_data, self)

                except Exception as ex:
                    logger.exception(ex)
                    continue
        except TypeError:
            # If we got a type error, that means retry was none, so we should end the thread
            return

    def stop(self):
        """
        Stop monitoring the account.

        The thread will terminate in up to X seconds, where X is the timeout set by the blockchain.
        """

        # Set the stop event, this will terminate the thread once we get the next event from the blockchain.
        self.stop_event.set()

        # Change the retry value,
        # this will cause an exception when trying to reconnect after timeout by the blockchain,
        # which will terminate the thread.
        self.sse_client.retry = None


class MultiMonitor:
    """Multi Monitor to monitor Kin payment on a multiple accounts"""

    def __init__(self, kin_client, addresses, callback_fn):
        """
        Monitors multiple accounts for kin payments
        :param kin_client: a kin client directed to the correct network
        :param addresses: addresses to watch
        :param callback_fn: function to callback when the payment is found
        """
        self.kin_client = kin_client
        self.callback_fn = callback_fn

        if not addresses:
            raise ValueError('no address to monitor')

        for address in addresses:
            if not is_valid_address(address):
                raise StellarAddressInvalidError('invalid address: {}'.format(address))
            if not self.kin_client.does_account_exists(address):
                raise AccountNotFoundError(address)

        self.addresses = addresses

        # Currently, due to nonstandard SSE implementation in Horizon,
        # using cursor=now will block until the first tx happens.
        # Instead, we determine the cursor ourselves.
        # Fix will be for horizon to send any message just to start a connection
        params = {}
        reply = self.kin_client.horizon.transactions(params={'order': 'desc', 'limit': 1})
        if len(reply['_embedded']['records']) == 1:
            cursor = reply['_embedded']['records'][0]['paging_token']
            params = {'cursor': cursor}

        # make synchronous SSE request (will raise errors in the current thread)
        self.sse_client = self.kin_client.horizon.transactions(sse=True, params=params)

        self.stop_event = Event()
        # start monitoring thread
        self.thread = Thread(target=self.event_processor, args=(self.stop_event,))
        self.thread.daemon = True
        self.thread.start()

    def event_processor(self, stop_event):
        """
        Method to filter through SSE events and find kin payments for an account
        :param stop_event: an event that can be used to stop this method
        """
        import json
        try:
            for event in self.sse_client:
                if stop_event.is_set():
                    return
                if event.event != 'message':
                    continue
                try:
                    tx = json.loads(event.data)

                    try:
                        tx_data = SimplifiedTransaction(RawTransaction(tx))
                    except CantSimplifyError:
                        continue

                    if tx_data.operation.type != OperationTypes.PAYMENT:
                        continue

                    if tx_data.source in self.addresses:
                        self.callback_fn(tx_data.source, tx_data, self)
                    if tx_data.operation.destination in self.addresses:
                        self.callback_fn(tx_data.operation.destination, tx_data, self)

                except Exception as ex:
                    logger.exception(ex)
                    continue
        except TypeError:
            # If we got a type error, that means retry was none, so we should end the thread
            return

    def stop(self):
        """
        Stop monitoring the account.

        The thread will terminate in up to X seconds, where X is the timeout set by the blockchain.
        """

        # Set the stop event, this will terminate the thread once we get the next event from the blockchain.
        self.stop_event.set()

        # Change the retry value,
        # this will cause an exception when trying to reconnect after timeout by the blockchain,
        # which will terminate the thread.
        self.sse_client.retry = None

    def add_address(self, address):
        """
        Add address to the watched addresses list
        :param address: address to add
        """
        if address in self.addresses:
            return

        if self.stop_event.is_set():
            raise StoppedMonitorError()

        if not is_valid_address(address):
            raise StellarAddressInvalidError('invalid address: {}'.format(address))

        if not self.kin_client.does_account_exists(address):
            raise AccountNotFoundError(address)

        self.addresses.append(address)

    def remove_address(self, address):
        """
        Remove an address for the list of addresses to watch
        :param address: the address to remove
        """
        if self.stop_event.is_set():
            raise StoppedMonitorError()

        self.addresses.remove(address)



from .config import TEST_ENVIRONMENT, PROD_ENVIRONMENT, SDK_USER_AGENT
from .client import KinClient
from . import errors as KinErrors
from .transactions import OperationTypes, decode_transaction
from .blockchain.keypair import Keypair
from .blockchain.environment import Environment
